// Copyright 1992-3 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "book.h"
#include "movegen.h"
#include "moveord.h"
#include "util.h"
extern "C"
{
#include <stdlib.h>
#include <time.h>
};

Book::Book()
{
    time_t t;
    srand((unsigned)time(&t));
    bookb = new Book_Reader("#1");
    bookw = new Book_Reader("#2");
}       

Book::~Book()
{
    delete bookb;
    delete bookw;
    Book_Entry::freeAll(TRUE);
}

Move Book::pick( const Board &b, uint16 node_index,
   const Book_Entry &node )
{
     Book_Reader *reader = (b.Side() == White) ? bookw : bookb;
     unsigned total_weight = 0;
     uint16 indx = node_index;
     while (indx != INVALID)
     {
        Book_Entry be;
        reader->Fetch(indx, be);
        if (be.is_equal(node))
           total_weight += be.get_recommend();
        if (be.is_last())
           break;
        else
           indx++;
     }
     if (total_weight == 0)
        return Move::NullMove();
     unsigned weight = 0;
     indx = node_index;
     const unsigned nRand = (rand()*(unsigned long)total_weight)/RAND_MAX;
     while (indx != INVALID)
     {
        Book_Entry be;
        reader->Fetch(indx,be);
        if (be.is_equal(node) && be.get_recommend() > 0)
        {
           weight += be.get_recommend();
           if (nRand < weight)
           {
               Move moves[Move_Generator::MaxMoves];
               Move_Generator mg( b, 0, Move::NullMove() );
               int n = mg.Generate_Moves(moves,False,True); 
               ASSERT(be.move_index < n);
               return moves[be.move_index];
           }
        }
        if (be.is_last())
           break;
        else
           indx++;
     }
     // should never get here
     ASSERT(False);
     return Move::NullMove();
}

Move Book::book_move( const Board &b )
{
     Book_Reader *reader = (b.Side() == White) ? bookw : bookb;
     if (!reader->Is_Open())
     {
         return Move::NullMove();
     }
     else
     {
        uint16 indx = reader->Head(b);
        Book_Entry target(b.HashCode(),0,0,0);
        return pick(b, indx, target);
     }
     return Move::NullMove();
}

unsigned Book::book_moves( const Board &b, Move *moves, 
        const unsigned limit)
{
     Book_Reader *reader = (b.Side() == White) ? bookw : bookb;
     if (!reader->Is_Open())
         return 0;

     unsigned indx = reader->Head(b);
     Book_Entry target(b.HashCode(),0,0,0);
     int num_moves = 0;
     while (indx != INVALID)
     {
        Book_Entry be;
        reader->Fetch(indx,be);
        if (target.is_equal(be) && be.get_recommend())
            num_moves++;
        if (be.is_last())
            break;
        else
            indx++;
    }
    if (num_moves == 0)
        return 0;
    Move *tmp_moves = new Move[Move_Generator::MaxMoves];
    Move_Generator mg( b, 0, Move::NullMove() );
    int n = mg.Generate_Moves(tmp_moves,False,True); 
    int * scores = new int[num_moves];
    Move * tmp_moves2 = new Move[num_moves];
    indx = reader->Head(b);
    num_moves = 0;
    while (indx != INVALID)
    {
        Book_Entry be;
        reader->Fetch(indx,be);
        if (target.is_equal(be) && be.get_recommend())
        {
           ASSERT(be.move_index < n);
           tmp_moves2[num_moves] = tmp_moves[be.move_index];
           scores[num_moves] = be.get_recommend();
           ++num_moves;
        }
        if (be.is_last())
           break;
        else
           indx++;
    }
    Move_Ordering::sort_moves(tmp_moves2,scores,num_moves);
    unsigned ret_val = Util::Min(num_moves,limit);
    for (unsigned i = 0; i < ret_val; i++)
       moves[i] = tmp_moves2[i];
    delete [] scores;
    delete [] tmp_moves;
    delete [] tmp_moves2;
    return ret_val;
}
